
# Number of records to bulk fetch/insert per batch for bootstrap operations.
#
# Benchmark testing of this value was performed by bootstrapping a table with
# five (5) columns and ~2.6 million rows. Two benchmark runs were performed with
# the database reset and restarted between runs. The first benchmark used
# 'batch_size=1000' and completed in 6min 15sec, the second benchmark used
# 'batch_size=10000' and completed in 6min 12sec (less than 1% difference in
# runtime). There was no noticeable difference in database resource usage
# between the two runs, but the Django process consistently used about 160MiB
# more memory for the duration of the second benchmark compared to the first.
# Given that the overall runtime was relatively unaffected between two tests
# whose batch size differed by an order of magnitude, the lower value seems like
# the better default due to the lower Django resource usage.
#
# Installations with noticeable database connection latency may prefer to
# specify a higher value on their bootstrap operations in order to optimize for
# fewer round-trips to the database.
BOOTSTRAP_BATCH_SIZE = 1000



import contextvars
from functools import wraps

from django.db import models, router, transaction

from .utils import get_fqcn

__all__ = [
    "AlreadyAudited",
    "audit_fields",
    "get_audited_class_path",
    "get_audited_models",
    "request",
]


class AlreadyAudited(Exception):
    """Model class is already audited."""


class InvalidManagerError(Exception):
    """Model class has an invalid manager."""


def audit_fields(*field_names, class_path=None, audit_special_queryset_writes=False):  # noqa: E501
    """Class decorator for auditing field changes on DB model instances.

    Use this on Model subclasses which need field change auditing.

    :param field_names: names of fields on the model that need to be audited
    :param class_path: optional name to use as the ``object_class_path`` field
        on audit events. The default (``None``) means the audited model's
        fully qualified "dot path" will be used.
    :param audit_special_queryset_writes: optional bool (default ``False``)
        enables auditing of the "special" QuerySet write methods (see below) for
        this model. Setting this to ``True`` requires the decorated model's
        default manager be an instance of ``field_audit.models.AuditingManager``
        **IMPORTANT**: These special QuerySet write methods often perform bulk
        or batched operations, and auditing their changes may impact efficiency.
    :raises: ``ValueError``, ``InvalidManagerError``

    Auditing is performed by decorating the model class's ``__init__()``,
    ``delete()`` and ``save()`` methods which provides audit events for all DB
    write operations except the "special" QuerySet write methods:

    - ``QuerySet.bulk_create()``
    - ``QuerySet.bulk_update()``
    - ``QuerySet.update()``
    - ``QuerySet.delete()``

    Using ``audit_special_queryset_writes=True`` (with the custom manager) lifts
    this limitation.
    """
    def wrapper(cls):
        if cls in _audited_models:
            raise AlreadyAudited(cls)
        if not issubclass(cls, models.Model):
            raise ValueError(f"expected Model subclass, got: {cls}")
        AuditEvent.attach_field_names(cls, field_names)
        if audit_special_queryset_writes:
            _verify_auditing_manager(cls)
        cls.__init__ = _decorate_init(cls.__init__)
        cls.save = _decorate_db_write(cls.save)
        cls.delete = _decorate_db_write(cls.delete)
        cls.refresh_from_db = _decorate_refresh_from_db(cls.refresh_from_db)
        _audited_models[cls] = get_fqcn(cls) if class_path is None else class_path  # noqa: E501
        return cls
    if not field_names:
        raise ValueError("at least one field name is required")
    from .models import AuditEvent
    return wrapper


def _verify_auditing_manager(cls):
    """Verifies a model class is configured with an appropriate manager for
    special QuerySet write auditing.

    :param cls: a Model subclass
    :raises: ``InvalidManagerError``
    """
    from .models import AuditingManager
    # Don't assume 'cls.objects', use 'cls._default_manager' instead.
    # see: https://docs.djangoproject.com/en/4.0/topics/db/managers/#default-managers  # noqa: E501
    if not isinstance(cls._default_manager, AuditingManager):
        raise InvalidManagerError(
            "QuerySet write auditing requires an AuditingManager, got "
            f"{type(cls._default_manager)}"
        )


def _decorate_init(init):
    """Decorates the "initialization" (e.g. __init__) method on Model
    subclasses. Responsible for ensuring that the initial field values are
    recorded in order to generate an audit event change delta later.

    :param init: the  method to decorate
    """
    @wraps(init)
    def wrapper(self, *args, **kw):
        init(self, *args, **kw)
        AuditEvent.attach_initial_values(self)
    from .models import AuditEvent
    return wrapper


def _decorate_db_write(func):
    """Decorates the "database write" methods (e.g. save, delete) on Model
    subclasses. Responsible for creating an audit event when a model instance
    changes.

    :param func: the "db write" method to decorate
    """
    @wraps(func)
    def wrapper(self, *args, **kw):
        # for details on using 'self._state', see:
        # - https://docs.djangoproject.com/en/dev/ref/models/instances/#state
        # - https://stackoverflow.com/questions/907695/
        is_create = is_save and self._state.adding
        object_pk = self.pk if is_delete else None

        db = router.db_for_write(type(self))
        with transaction.atomic(using=db):
            ret = func(self, *args, **kw)
            AuditEvent.audit_field_changes(
                self,
                is_create,
                is_delete,
                request.get(),
                object_pk,
            )
            return ret
    is_save = func.__name__ == "save"
    is_delete = func.__name__ == "delete"
    if not is_save and not is_delete:
        raise ValueError(f"invalid function for decoration: {func}")
    from .models import AuditEvent
    return wrapper


def _decorate_refresh_from_db(func):
    """Decorates the "refresh from db" method on Model subclasses. This is
    necessary to ensure that all audited fields are included in the refresh
    to avoid recursively calling the refresh for deferred fields.

    :param func: the "refresh from db" method to decorate
    """
    @wraps(func)
    def wrapper(self, using=None, fields=None, **kwargs):
        if fields is not None:
            fields = set(fields) | set(AuditEvent.field_names(self))
        func(self, using, fields, **kwargs)

    from .models import AuditEvent
    return wrapper


def get_audited_models():
    return _audited_models.copy()


def get_audited_class_path(cls):
    return _audited_models[cls]


_audited_models = {}
request = contextvars.ContextVar("request", default=None)



from .field_audit import request as audit_request


class FieldAuditMiddleware:
    """Middleware that gives the ``audit_fields()`` decorator access to the
    Django request.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def process_view(self, request, view_func, view_args, view_kwargs):
        audit_request.set(request)
        return None

    def __call__(self, request):
        return self.get_response(request)



from django.db import migrations, models

from ..models import get_date


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='AuditEvent',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),  # noqa: E501
                ('event_date', models.DateTimeField(db_index=True, default=get_date)),  # noqa: E501
                ('object_class_path', models.CharField(db_index=True, max_length=255)),  # noqa: E501
                ('object_pk', models.JSONField()),
                ('change_context', models.JSONField()),
                ('is_create', models.BooleanField(default=False)),
                ('is_delete', models.BooleanField(default=False)),
                ('delta', models.JSONField()),
            ],
        ),
        migrations.AddConstraint(
            model_name='auditevent',
            constraint=models.CheckConstraint(
                name='field_audit_auditevent_valid_create_or_delete',
                check=models.Q(
                    ('is_create', True),
                    ('is_delete', True),
                    _negated=True,
                ),
            ),
        ),
    ]



import logging

from django.core.exceptions import ImproperlyConfigured
from django.db.migrations import RunPython
from django.utils.module_loading import import_string

from .const import BOOTSTRAP_BATCH_SIZE

log = logging.getLogger(__name__)


def class_import_helper(dotted_path, item_description, require_type=None):
    """Returns an imported class described by ``dotted_path``.

    Similar to ``django.db.models.options.Options._get_default_pk_class()``.
    It would be nice if Django had a helper function for doing this.

    :param dotted_path: Python syntax "dotted path" of class to be imported.
    :param item_description: Description for generating helpful exception
                             messages.
    :param required_type: (Optional) require that the imported class is a
                          subclass of this.
    :raises: ImproperlyConfigured, ValueError
    """
    if not isinstance(dotted_path, str):
        # Django should implement this check to avoid confusing errors
        # from `import_string()`, like:
        # AttributeError: type object 'AutoField' has no attribute 'rsplit'
        raise ValueError(
            f"invalid {item_description}: expected 'str', "
            f"got {dotted_path!r}"
        )

    try:
        class_ = import_string(dotted_path)
    except ImportError as exc:
        msg = f"failed to import {item_description}: {dotted_path!r}"
        raise ImproperlyConfigured(msg) from exc

    if require_type is not None:
        if not issubclass(class_, require_type):
            raise ValueError(
                f"invalid imported {item_description}: expected subclass of "
                f"{require_type.__name__!r}, got {class_!r}"
            )
    return class_


def get_fqcn(cls):
    """Get the full dot-delimited class path (module + qualname)

    See: https://stackoverflow.com/a/2020083
    """
    return f"{cls.__module__}.{cls.__qualname__}"


def run_bootstrap(model_class, field_names, batch_size=BOOTSTRAP_BATCH_SIZE,
                  iter_records=None, reverse_func=RunPython.noop):
    """Returns a django migration Operation which calls
    ``AuditEvent.bootstrap_existing_model_records()`` to add "migration" records
    for existing model records.

    :param model_class: see
        ``field_audit.models.AuditEvent.bootstrap_existing_model_records``
    :param field_names: see
        ``field_audit.models.AuditEvent.bootstrap_existing_model_records``
    :param batch_size: see
        ``field_audit.models.AuditEvent.bootstrap_existing_model_records``,
        (default=field_audit.const.BOOTSTRAP_BATCH_SIZE)
    :param iter_records:  see
        ``field_audit.models.AuditEvent.bootstrap_existing_model_records``
    :param reverse_func: (optional, default: ``RunPython.noop``) a callable for
        unapplying the migration. Passed directly to the returned
        ``RunPython()`` instance as the ``reverse_code`` argument.
    """
    def do_bootstrap(*args, **kwargs):
        from .models import AuditEvent
        count = AuditEvent.bootstrap_existing_model_records(
            model_class,
            field_names,
            batch_size,
            iter_records,
        )
        log.info(
            f"bootstrapped {count} audit event{'' if count == 1 else 's'} for: "
            f"{model_class._meta.app_label}.{model_class._meta.object_name}"
        )
    return RunPython(do_bootstrap, reverse_code=reverse_func)



from enum import Enum
from functools import wraps
from itertools import islice

from django.conf import settings
from django.db import models, transaction
from django.db.models import Expression
from django.utils import timezone

from .const import BOOTSTRAP_BATCH_SIZE
from .utils import class_import_helper

USER_TYPE_TTY = "SystemTtyOwner",
USER_TYPE_PROCESS = "SystemProcessOwner"
USER_TYPE_REQUEST = "RequestUser"


def check_engine_sqlite(engine=None):
    """Check if SQLite (or Oracle) database engines are in use. If ``engine`` is
    ``None``, check all Django engines (otherwise check only ``engine``).

    :param engine: (Optional) name of a Django database engine.
    """
    def lite_it_up(engine):
        # check if engine "flavor" is Oracle or SQLite
        # example db engine: django.db.backends.sqlite3
        # resulting flavor:                     sqlite
        return engine.split(".")[-1][:6] in {"sqlite", "oracle"}
    if engine is None:
        for db_properties in settings.DATABASES.values():
            # The following "no branch" directive prevents coverage from
            # reporting the (known) untested branch of
            # `if lite_it_up(...)->return`. There will always be an uncovered
            # branch here because tests only run on postgres _or_ sqlite, never
            # both.
            if lite_it_up(db_properties["ENGINE"]):  # pragma: no branch
                return True
    else:
        return lite_it_up(engine)
    # "no cover" note: tests only run on postgres _or_ sqlite, never both
    return False  # pragma: no cover


class AuditEventManager(models.Manager):
    """Manager for the AuditEvent model."""

    def by_model(self, model_class):
        """Filter records for a specific model.

        :param model_class: an audited Django model class
        :returns: ``QuerySet``
        """
        from .field_audit import get_audited_class_path
        return self.filter(
            object_class_path=get_audited_class_path(model_class)
        )

    def cast_object_pk_for_model(self, model_class):
        """Filter records for a specific model and add an ``as_pk_type``
        expression column containing the ``object_pk`` values cast to the PK
        type of ``model_class``.

        :param model_class: an audited Django model class
        :returns: ``QuerySet``
        """
        if type(model_class._meta.pk) is models.JSONField:
            expression = models.F("object_pk")
        else:
            expression = CastFromJson("object_pk", model_class._meta.pk)
        return self.by_model(model_class).annotate(as_pk_type=expression)

    def cast_object_pks_list(self, model_class):
        """Convenience method for getting the results of
        ``cast_object_pk_for_model(...)`` as a values list.

        Example:
        >>> SomeModel.objects.filter(pk_in=(
            AuditEvent.objects
            .filter(event_date__gte=datetime.date.today())
            .cast_object_pks_list(SomeModel)
        ))

        :param model_class: an audited Django model class
        :param flat: optional argument passed to the
            ``values_list()`` method (default=True).
        """
        return (
            self.cast_object_pk_for_model(model_class)
            .values_list("as_pk_type", flat=True)
        )

    def by_type_and_username(self, user_type, username):
        """Use the ``contains`` query (PostgreSQL and MySQL/MariaDB only) to
        query for documents with matching keys.

        If other DB flavors are configured in Django settings, support is
        defined at import time (see below)
        """
        # "no cover" note: tests only run on postgres _or_ sqlite, never both
        return self.filter(  # pragma: no cover
            change_context__contains={
                "user_type": user_type,
                "username": username,
            },
        )

    # "no branch" note: tests only run on postgres _or_ sqlite, never both
    if check_engine_sqlite():  # pragma: no branch
        _by_type_and_username = by_type_and_username

        def by_type_and_username(self, user_type, username):
            """Support SQLite (for development) and Oracle (because it comes
            along for free).

            If these DB flavors are not needed at import time, this "extra db"
            support is never defined.
            """
            if check_engine_sqlite(settings.DATABASES[self.db]["ENGINE"]):
                # Oracle and SQLite do not support `contains` queries
                # see: https://docs.djangoproject.com/en/4.0/topics/db/queries/#std:fieldlookup-jsonfield.contains  # noqa: E501
                return self.filter(
                    change_context__user_type=user_type,
                    change_context__username=username,
                )
            # "no cover" note: tests only run on postgres _or_ sqlite, never
            # both
            return self._by_type_and_username(user_type, username)  # pragma: no cover  # noqa: E501


class CastFromJson(models.functions.comparison.Cast):

    def __init__(self, expression, output_field):
        super().__init__(JsonPreCast(expression), output_field)


class JsonPreCast(models.expressions.Func):
    """A function that works on the JSON type and prepares it such that it can
    be cast to other types."""

    template = "%(expressions)s"
    arity = 1

    def as_postgresql(self, compiler, connection, **extra_context):
        sql, params = self.as_sql(compiler, connection, **extra_context)
        return f"({sql} #>> '{{}}')", params

    def as_sqlite(self, compiler, connection, **extra_context):
        sql, params = self.as_sql(compiler, connection, **extra_context)
        return f"JSON_EXTRACT({sql}, '$')", params


class DefaultAuditEventManager(AuditEventManager):
    """Default Manager for the AuditEvent model. Contains convenience methods
    for the default auditors, which may not be desirable to subclass if
    downstream projects wish to define custom auditor chains.
    """

    def by_system_user(self, username):
        system_types = [USER_TYPE_TTY, USER_TYPE_PROCESS]
        return self.filter(
            change_context__user_type__in=system_types,
            change_context__username=username,
        )

    def by_tty_user(self, username):
        return self.by_type_and_username(USER_TYPE_TTY, username)

    def by_process_user(self, username):
        return self.by_type_and_username(USER_TYPE_PROCESS, username)

    def by_request_user(self, username):
        return self.by_type_and_username(USER_TYPE_REQUEST, username)


def get_manager(attr_suffix, default):
    """Returns an instantiated manager, possibly one defined in settings.

    If the manager is defined in settings, it must be a subclass of
    ``django.db.models.Manager``.

    :param attr_suffix: Suffix (appended to ``FIELD_AUDIT_``) of settings
        attribute for the manager class path.
    :param default: Manager class. Returned if attribute isn't defined.
    """
    settings_attr = f"FIELD_AUDIT_{attr_suffix}"
    try:
        class_path = getattr(settings, settings_attr)
    except AttributeError:
        return default()
    desc = f"{settings_attr!r} value"
    return class_import_helper(class_path, desc, models.Manager)()


def get_date():
    """Returns the current UTC date/time.

    This is the "getter" for default values of the ``AuditEvent.event_date``
    field.
    """
    return timezone.now()


class AuditEvent(models.Model):
    event_date = models.DateTimeField(default=get_date, db_index=True)
    object_class_path = models.CharField(db_index=True, max_length=255)
    object_pk = models.JSONField()
    change_context = models.JSONField()
    is_create = models.BooleanField(default=False)
    is_delete = models.BooleanField(default=False)
    is_bootstrap = models.BooleanField(default=False)
    delta = models.JSONField()

    objects = get_manager("AUDITEVENT_MANAGER", DefaultAuditEventManager)

    class Meta:
        constraints = [
            models.CheckConstraint(
                name="field_audit_auditevent_chk_create_or_delete_or_bootstrap",
                check=~(
                    models.Q(is_create=True, is_delete=True) | \
                    models.Q(is_create=True, is_bootstrap=True) | \
                    models.Q(is_delete=True, is_bootstrap=True)  # noqa: E502
                ),
            ),
        ]

    ATTACH_FIELD_NAMES_AT = "__field_audit_field_names"
    ATTACH_INIT_VALUES_AT = "__field_audit_init_values"

    @classmethod
    def attach_field_names(cls, model_class, field_names):
        """Attaches a collection of field names to a Model class for auditing.

        :param model_class: a Django Model class under audit
        :param field_names: collection of field names to audit on the model
        """
        setattr(model_class, cls.ATTACH_FIELD_NAMES_AT, field_names)

    @classmethod
    def field_names(cls, model_class):
        """Returns the audit field names stored on the audited Model class

        :param model_class: a Django Model class under audit
        """
        return getattr(model_class, cls.ATTACH_FIELD_NAMES_AT)

    @staticmethod
    def get_field_value(instance, field_name):
        """Returns the database value of a field on ``instance``.

        :param instance: an instance of a Django model
        :param field_name: name of a field on ``instance``
        """
        field = instance._meta.get_field(field_name)
        return field.to_python(field.value_from_object(instance))

    @classmethod
    def attach_initial_values(cls, instance):
        """Save copies of field values on an instance so they can be used later
        to determine if the instance has changed and record what the previous
        values were.

        :param instance: instance of a Model subclass to be audited for changes
        :raises: ``AttachValuesError`` if initial values are already attached to
            the instance
        """
        if hasattr(instance, cls.ATTACH_INIT_VALUES_AT):
            # This should never happen, but to be safe, refuse to clobber
            # existing attributes.
            raise AttachValuesError(
                f"refusing to overwrite {cls.ATTACH_INIT_VALUES_AT!r} "
                f"on model instance: {instance}"
            )
        field_names = cls.field_names(instance)
        init_values = {f: cls.get_field_value(instance, f) for f in field_names}
        setattr(instance, cls.ATTACH_INIT_VALUES_AT, init_values)

    @classmethod
    def reset_initial_values(cls, instance):
        """Returns the previously attached "initial values" and attaches new
        values.

        :param instance: instance of a Model subclass to be audited for changes
        :raises: ``AttachValuesError`` if initial values are not attached to
            the instance
        """
        try:
            values = getattr(instance, cls.ATTACH_INIT_VALUES_AT)
        except AttributeError:
            raise AttachValuesError("cannot reset values that were never set")
        delattr(instance, cls.ATTACH_INIT_VALUES_AT)
        cls.attach_initial_values(instance)
        return values

    @classmethod
    def audit_field_changes(cls, *args, **kw):
        """Convenience method that calls ``make_audit_event_from_instance()``
        and saves the event (if one is returned).

        All [keyword] arguments are passed directly to
        ``make_audit_event_from_instance()``, see that method for usage.
        """
        event = cls.make_audit_event_from_instance(*args, **kw)
        if event is not None:
            event.save()

    @classmethod
    def get_delta_from_instance(cls, instance, is_create, is_delete):
        """
        Returns a dictionary representing the delta of an instance of a model
        being audited for changes.
        Has the side effect of calling cls.reset_initial_values(instance)
        which grabs and updates the initial values stored on the instance.

        :param instance: instance of a Model subclass to be audited for changes
        :param is_create: whether or not the audited event creates a new DB
            record (setting ``True`` implies that ``instance`` is changing)
        :param is_delete: whether or not the audited event deletes an existing
            DB record (setting ``True`` implies that ``instance`` is changing)
        :returns: {field_name: {'old': old_value, 'new': new_value}, ...}
        :raises: ``AssertionError`` if both is_create and is_delete are true
        """
        assert not (is_create and is_delete), \
            "is_create and is_delete cannot both be true"
        fields_to_audit = cls.field_names(instance)
        # SIDE EFFECT: fetch and reset initial values for next db write
        init_values = cls.reset_initial_values(instance)
        old_values = {} if is_create else init_values
        new_values = {} if is_delete else \
            {f: cls.get_field_value(instance, f) for f in fields_to_audit}
        return cls.create_delta(old_values, new_values)

    @staticmethod
    def create_delta(old_values, new_values):
        """
        Compares two dictionaries and creates a delta between the two

        :param old_values: {field_name: field_value, ...} representing the
        values prior to a change
        :param new_values: {field_name: field_value, ...} representing the
        values after a change
        :returns: {field_name: {'old': old_value, 'new': new_value}, ...}
        :raises: ``AssertionError`` if both old_values and new_values are empty
        do not match
        """
        assert old_values or new_values, \
            "Must provide a non-empty value for either old_values or new_values"

        changed_fields = old_values.keys() if old_values else new_values.keys()
        if old_values and new_values:
            changed_fields = new_values.keys()

        delta = {}
        for field_name in changed_fields:
            if not old_values:
                delta[field_name] = {"new": new_values[field_name]}
            elif not new_values:
                delta[field_name] = {"old": old_values[field_name]}
            else:
                try:
                    old_value = old_values[field_name]
                except KeyError:
                    delta[field_name] = {"new": new_values[field_name]}
                else:
                    if old_value != new_values[field_name]:
                        delta[field_name] = {"old": old_value,
                                             "new": new_values[field_name]}
        return delta

    @classmethod
    def make_audit_event_from_instance(cls, instance, is_create, is_delete,
                                       request, object_pk=None):
        """Factory method for creating a new ``AuditEvent`` for an instance of a
        model that's being audited for changes.

        :param instance: instance of a Model subclass to be audited for changes
        :param is_create: whether or not the audited event creates a new DB
            record (setting ``True`` implies that ``instance`` is changing)
        :param is_delete: whether or not the audited event deletes an existing
            DB record (setting ``True`` implies that ``instance`` is changing)
        :param request: the request object responsible for the change (or
            ``None`` if there is no request)
        :param object_pk: (Optional) primary key of the instance. Only used when
            ``is_delete == True``, that is, when the instance itself no longer
            references its pre-delete primary key. It is ambiguous to set this
            when ``is_delete == False``, and doing so will raise an exception.
        :returns: an unsaved ``AuditEvent`` instance (or ``None`` if
            ``instance`` has not changed)
        :raises: ``ValueError`` on invalid use of the ``object_pk`` argument
        """
        if not is_delete:
            if object_pk is not None:
                raise ValueError(
                    "'object_pk' arg is ambiguous when 'is_delete == False'"
                )
            object_pk = instance.pk

        delta = cls.get_delta_from_instance(instance, is_create, is_delete)

        if delta:
            return cls.create_audit_event(object_pk, type(instance), delta,
                                          is_create, is_delete, request)

    @classmethod
    def make_audit_event_from_values(cls, old_values, new_values, object_pk,
                                     object_cls, request):
        """Factory method for creating a new ``AuditEvent`` based on old and new
        values.

        :param old_values: {field_name: field_value, ...} representing the
        values prior to a change
        :param new_values: {field_name: field_value, ...} representing the
        values after a change
        :param object_pk: primary key of the instance
        :param object_cls: class type of the object being audited
        :param request: the request object responsible for the change (or
            ``None`` if there is no request)
        :returns: an unsaved ``AuditEvent`` instance (or ``None`` if
            no difference between ``old_values`` and ``new_values``)
        """
        is_create = not old_values
        is_delete = not new_values
        delta = AuditEvent.create_delta(old_values, new_values)
        if delta:
            return AuditEvent.create_audit_event(object_pk, object_cls, delta,
                                                 is_create, is_delete, request)

    @classmethod
    def create_audit_event(cls, object_pk, object_cls, delta, is_create,
                           is_delete, request):
        from .auditors import audit_dispatcher
        from .field_audit import get_audited_class_path
        change_context = audit_dispatcher.dispatch(request)
        object_cls_path = get_audited_class_path(object_cls)
        return cls(
            object_class_path=object_cls_path,
            object_pk=object_pk,
            change_context=cls._change_context_db_value(change_context),
            is_create=is_create,
            is_delete=is_delete,
            delta=delta,
        )

    @classmethod
    def bootstrap_existing_model_records(cls, model_class, field_names,
                                         batch_size=BOOTSTRAP_BATCH_SIZE,
                                         iter_records=None):
        """Creates audit events for all existing records of ``model_class``.
        Database records are fetched and created in batched bulk operations
        for efficiency.

        :param model_class: a subclass of ``django.db.models.Model`` that uses
            the ``audit_fields()`` decorator.
        :param field_names: a collection of field names to include in the
            resulting audit event ``delta`` value.
        :param batch_size: (optional) create bootstrap records in batches of
            ``batch_size``. Default: ``field_audit.const.BOOTSTRAP_BATCH_SIZE``.
            Use ``None`` to disable batching.
        :param iter_records: a callable used to fetch model instances.
            If ``None`` (the default), ``.all().iterator()`` is called on the
            model's default manager.
        :returns: number of bootstrap records created
        """
        from .auditors import audit_dispatcher
        from .field_audit import get_audited_class_path

        if iter_records is None:
            iter_records = model_class._default_manager.all().iterator

        def iter_events():
            for instance in iter_records():
                delta = {}
                for field_name in field_names:
                    value = cls.get_field_value(instance, field_name)
                    delta[field_name] = {"new": value}
                yield cls(
                    object_class_path=object_class_path,
                    object_pk=instance.pk,
                    change_context=change_context,
                    is_bootstrap=True,
                    delta=delta,
                )

        change_context = cls._change_context_db_value(
            audit_dispatcher.dispatch(None)
        )
        object_class_path = get_audited_class_path(model_class)

        if batch_size is None:
            return len(cls.objects.bulk_create(iter_events()))
        # bulk_create in batches efficiently
        # see: https://docs.djangoproject.com/en/4.0/ref/models/querysets/#bulk-create  # noqa: E501
        events = iter_events()
        total = 0
        while True:
            batch = list(islice(events, batch_size))
            if not batch:
                break
            total += len(batch)
            cls.objects.bulk_create(batch, batch_size=batch_size)
        return total

    @classmethod
    def bootstrap_top_up(cls, model_class, field_names,
                         batch_size=BOOTSTRAP_BATCH_SIZE):
        """Creates audit events for existing records of ``model_class`` which
        were created prior to auditing being enabled and are lacking a bootstrap
        or create AuditEvent record.

        :param model_class: see ``bootstrap_existing_model_records``
        :param field_names: see ``bootstrap_existing_model_records``
        :param batch_size:  see ``bootstrap_existing_model_records``
            (default=field_audit.const.BOOTSTRAP_BATCH_SIZE)
        :returns: number of bootstrap records created
        """
        subquery = (
            cls.objects
            .cast_object_pks_list(model_class)
            .filter(
                models.Q(models.Q(is_bootstrap=True) | models.Q(is_create=True))
            )
        )
        # bootstrap the model records who do not match the subquery
        model_manager = model_class._default_manager
        return cls.bootstrap_existing_model_records(
            model_class,
            field_names,
            batch_size=batch_size,
            iter_records=model_manager.exclude(pk__in=subquery).iterator,
        )

    @classmethod
    def _change_context_db_value(cls, value):
        return {} if value is None else value

    def __repr__(self):  # pragma: no cover
        cls_name = type(self).__name__
        return f"<{cls_name} ({self.id}, {self.object_class_path!r})>"


class AttachValuesError(Exception):
    """Attaching initial values to a Model instance failed."""


class InvalidAuditActionError(Exception):
    """A special QuerySet write method was called with non-AuditAction enum."""


class UnsetAuditActionError(Exception):
    """A special QuerySet write method was called without an audit action."""


class AuditAction(Enum):

    AUDIT = object()
    IGNORE = object()
    RAISE = object()

    def __repr__(self):  # pragma: no cover
        return f"<{type(self).__name__}.{self.name}>"


def validate_audit_action(func):
    """Decorator that performs validation on the ``audit_action`` keyword arg.

    :raises: ``InvalidAuditActionError`` or ``UnsetAuditActionError``
    """
    @wraps(func)
    def wrapper(self, *args, **kw):
        try:
            audit_action = kw["audit_action"]
        except KeyError:
            raise UnsetAuditActionError(
                f"{type(self).__name__}.{func.__name__}() requires an audit "
                "action as a keyword argument."
            )
        if audit_action not in AuditAction:
            raise InvalidAuditActionError(
                "The 'audit_action' argument must be a value of 'AuditAction', "
                f"got {type(audit_action)!r}"
            )
        if audit_action is AuditAction.RAISE:
            raise UnsetAuditActionError(
                f"{type(self).__name__}.{func.__name__}() requires an audit "
                "action"
            )
        return func(self, *args, **kw)
    return wrapper


class AuditingQuerySet(models.QuerySet):
    """A QuerySet that can perform field change auditing for bulk write methods.

    When decorating a model class with
    ``@audit_fields(..., audit_special_queryset_writes=True)``, the model's
    default manager must be a subclass of ``AuditingManager``.  Doing so
    provides the required extra auditing logic for the following methods:

    - ``bulk_create()``
    - ``bulk_update()``
    - ``delete()``
    - ``update()``

    Each of these methods has an additional required keyword argument
    ``audit_action`` which defaults to ``AuditAction.RAISE``.  When calling one
    of the above methods, the value must be explicitly set to one of:

    - ``AuditAction.AUDIT`` -- perform the database write, creating audit events
        for the changes.
    - ``AuditAction.IGNORE`` -- perform the database write without performing
        any auditing logic (pass through).

    Calling one of these methods without setting the desired audit action will
    raise an exception.
    """

    @validate_audit_action
    def bulk_create(self, objs, *, audit_action=AuditAction.RAISE, **kw):
        if audit_action is AuditAction.IGNORE or not objs:
            return super().bulk_create(objs, **kw)
        assert audit_action is AuditAction.AUDIT, audit_action

        from .field_audit import request
        request = request.get()

        with transaction.atomic(using=self.db):
            created_objs = super().bulk_create(objs, **kw)
            audit_events = []
            for obj in created_objs:
                audit_events.append(
                    AuditEvent.make_audit_event_from_instance(
                        obj, True, False, request))
            AuditEvent.objects.bulk_create(audit_events)
            return created_objs

    @validate_audit_action
    def bulk_update(self, *args, audit_action=AuditAction.RAISE, **kw):
        if audit_action is AuditAction.IGNORE:
            return super().bulk_update(*args, **kw)
        else:
            raise NotImplementedError(
                "Change auditing is not implemented for bulk_update()."
            )

    @validate_audit_action
    def delete(self, *, audit_action=AuditAction.RAISE):
        if audit_action is AuditAction.IGNORE:
            return super().delete()
        assert audit_action is AuditAction.AUDIT, audit_action
        from .field_audit import request
        request = request.get()
        audit_events = []
        fields_to_fetch = set(AuditEvent.field_names(self.model)) | {'pk'}
        current_values = {}
        for values_for_instance in self.values(*fields_to_fetch):
            pk = values_for_instance.pop('pk')
            current_values[pk] = values_for_instance

        for pk, current_values_for_pk in current_values.items():
            audit_event = AuditEvent.make_audit_event_from_values(
                current_values_for_pk,
                {},
                pk,
                self.model,
                request
            )
            audit_events.append(audit_event)

        with transaction.atomic(using=self.db):
            value = super().delete()
            if audit_events:
                # write the audit events _after_ the delete succeeds
                AuditEvent.objects.bulk_create(audit_events)
            return value

    @validate_audit_action
    def update(self, *, audit_action=AuditAction.RAISE, **kw):
        """
        In order to determine the old and new values of the records matched by
        the queryset, a fetch of audited values for the matched records is
        performed, resulting in one fetch of the current values, one update of
        the matched records, and one bulk creation of audit events.

        NOTE: if django.db.models.Expressions are provided as update arguments,
        an additional fetch of records is performed to obtain new values.
        """
        if audit_action is AuditAction.IGNORE:
            return super().update(**kw)
        assert audit_action is AuditAction.AUDIT, audit_action

        fields_to_update = set(kw.keys())
        audited_fields = set(AuditEvent.field_names(self.model))
        fields_to_audit = fields_to_update & audited_fields
        if not fields_to_audit:
            # no audited fields are changing
            return super().update(**kw)

        new_values = {field: kw[field] for field in fields_to_audit}
        uses_expressions = any(
            [isinstance(val, Expression) for val in new_values.values()])

        old_values = {}
        values_to_fetch = fields_to_update | {"pk"}
        for value in self.values(*values_to_fetch):
            pk = value.pop('pk')
            old_values[pk] = value

        with transaction.atomic(using=self.db):
            rows = super().update(**kw)
            if uses_expressions:
                # fetch updated values to ensure audit event deltas are accurate
                # after update is performed with expressions
                new_values = {}
                for value in self.values(*values_to_fetch):
                    pk = value.pop('pk')
                    new_values[pk] = value
            else:
                new_values = {pk: new_values for pk in old_values.keys()}

            # create and write the audit events _after_ the update succeeds
            from .field_audit import request
            request = request.get()
            audit_events = []
            for pk, old_values_for_pk in old_values.items():
                audit_event = AuditEvent.make_audit_event_from_values(
                    old_values_for_pk, new_values[pk], pk, self.model, request
                )
                if audit_event:
                    audit_events.append(audit_event)
            if audit_events:
                AuditEvent.objects.bulk_create(audit_events)
            return rows


AuditingManager = models.Manager.from_queryset(AuditingQuerySet)



import os
import re
from io import open

from setuptools import find_packages, setup


def get_version(filename):
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path, encoding="utf-8") as handle:
        content = handle.read()
    return re.search(r'__version__ = "([^"]+)"', content).group(1)


def read_md(filename):
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path, encoding="utf-8") as handle:
        return handle.read()


setup(
    name="django-field-audit",
    version=get_version("field_audit/__init__.py"),
    description="Audit Field Changes on Django Models",
    long_description=read_md("README.md"),
    long_description_content_type="text/markdown",
    maintainer="Joel Miller",
    maintainer_email="jmiller@dimagi.com",
    url="https://github.com/dimagi/django-field-audit",
    license="BSD License",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Framework :: Django",
        "Framework :: Django :: 3",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4",
        "Framework :: Django :: 4.2",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from django.apps import AppConfig


class FieldAuditConfig(AppConfig):

    name = "field_audit"
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self):
        from .auditors import audit_dispatcher
        audit_dispatcher.setup_auditors()



from getpass import getuser
from subprocess import DEVNULL, CalledProcessError, check_output

from .models import (
    USER_TYPE_PROCESS,
    USER_TYPE_REQUEST,
    USER_TYPE_TTY,
)
from .utils import class_import_helper

__all__ = [
    "BaseAuditor",
    "RequestAuditor",
    "SystemUserAuditor",
    "audit_dispatcher",
]


class _AuditDispatcher:
    """Dispatcher for the Audit API used to get "changed by" information when
    creating new ``AuditEvent`` records (i.e. when an audited model is changed).

    An instance of this class maintains the auditor chain that defines which
    (and in what order) ``BaseAuditor`` subclass instances are used to acquire
    "changed by" information.

    The auditor chain can be customized by defining a list of 'BaseAuditor'
    subclass paths via the ``FIELD_AUDIT_AUDITORS`` settings attribute.
    """

    def setup_auditors(self):
        """Populate the auditors chain, possibly defined in settings.

        This method is called at app ready time, and must be called before the
        ``dispatch()`` method can be used.
        """
        from django.conf import settings
        auditors_attr = "FIELD_AUDIT_AUDITORS"
        if hasattr(settings, auditors_attr):
            self.auditors = []
            for auditor_path in getattr(settings, auditors_attr):
                auditor_class = class_import_helper(
                    auditor_path,
                    f"{auditors_attr!r} item",
                    BaseAuditor,
                )
                self.auditors.append(auditor_class())
        else:
            self.auditors = [RequestAuditor(), SystemUserAuditor()]

    def dispatch(self, request):
        """Cycles through the auditors chain and returns the first non-None
        value returned by a call to ``auditor.change_context(request)``.

        :param request: Django request to be audited (or ``None``).
        :returns: JSON-serializable value (or ``None`` if chain is exhausted).
        """
        for auditor in self.auditors:
            change_context = auditor.change_context(request)
            if change_context is not None:
                return change_context
        return None


audit_dispatcher = _AuditDispatcher()


class BaseAuditor:
    """Abstract class for the Auditor API. Subclasses are used to return
    "changed by" information associated with an event which they know how to
    audit.

    BaseAuditor subclasses must define the following:
    - ``change_context()`` method that returns a JSON-serializable object of
      information for events it knows how to audit (or ``None`` otherwise).
    """

    def change_context(self, request):
        raise NotImplementedError("change_context() is abstract")


class RequestAuditor(BaseAuditor):
    """Auditor class for getting users from authenticated requests."""

    def change_context(self, request):
        if request is None:
            # cannot provide a request user without a request
            return None
        if request.user.is_authenticated:
            return {
                "user_type": USER_TYPE_REQUEST,
                "username": request.user.username,
            }
        # short-circuit the audit chain for not-None requests
        return {}


class SystemUserAuditor(BaseAuditor):
    """Auditor class for getting OS usernames."""

    def __init__(self):
        self.has_who_bin = True

    def change_context(self, request):
        username = None
        if self.has_who_bin:
            try:
                # get owner of STDIN file on login sessions (e.g. SSH)
                output = check_output(["who", "-m"], stderr=DEVNULL)
            except CalledProcessError:
                self.has_who_bin = False
            else:
                if output:
                    try:
                        username = output.split()[0].decode("utf-8")
                        user_type = USER_TYPE_TTY
                    except (IndexError, UnicodeDecodeError):
                        pass
        if not username:
            # no TTY user, get owner of the current process
            username = getuser()
            user_type = USER_TYPE_PROCESS
        if username:
            return {"user_type": user_type, "username": username}
        return None



from contextlib import contextmanager

from django.core.management.base import BaseCommand, CommandError

from field_audit.const import BOOTSTRAP_BATCH_SIZE
from field_audit.field_audit import get_audited_models
from field_audit.models import AuditEvent


class Command(BaseCommand):

    help = "Create bootstrap AuditEvent records for audited model classes."
    models = {}

    @classmethod
    def setup_models(cls):
        def model_name(model, unique=False):
            if unique:
                return f"{model._meta.app_label}.{model.__name__}"
            return model.__name__
        for model_class in get_audited_models():
            name = model_name(model_class)
            collision = cls.models.setdefault(name, model_class)
            if collision is not model_class:
                original = model_name(collision, unique=True)
                namesake = model_name(model_class, unique=True)
                if original == namesake:
                    raise InvalidModelState(
                        "Two audited models from the same app have the "
                        f"same name: {(collision, model_class)}"
                    )
                del cls.models[name]
                cls.models[original] = collision
                cls.models[namesake] = model_class

    def add_arguments(self, parser):
        parser.add_argument(
            "operation",
            choices=self.operations,
            help="Type of bootstrap operation to perform.",
        )
        parser.add_argument(
            "models",
            choices=sorted(self.models),
            nargs="+",
            help="Model class(es) to perform the bootstrap operation on.",
        )
        parser.add_argument(
            "--batch-size",
            default=BOOTSTRAP_BATCH_SIZE,
            metavar="N",
            type=int,
            help=(
                "Perform database queries in batches of size %(metavar)s "
                "(default=%(default)s). A value of zero (0) disables batching."
            ),
        )

    def handle(self, operation, models, batch_size, **options):
        self.stdout.ending = None
        self.logfile = self.stdout
        if batch_size < 0:
            raise CommandError("--batch-size must be a positive integer")
        if batch_size == 0:
            batch_size = None
        self.batch_size = batch_size
        for name in models:
            model_class = self.models[name]
            self.operations[operation](self, model_class)

    def init_all(self, model_class):
        query = model_class._default_manager.all()
        log_head = f"init: {model_class} ({query.count()}) ... "
        with self.bootstrap_action_log(log_head) as stream:
            count = self.do_bootstrap(
                model_class,
                AuditEvent.bootstrap_existing_model_records,
                iter_records=query.iterator,
            )
            stream.write(f"done ({count})")

    def top_up_missing(self, model_class):
        log_head = f"top-up: {model_class} ... "
        with self.bootstrap_action_log(log_head) as stream:
            count = self.do_bootstrap(model_class, AuditEvent.bootstrap_top_up)
            stream.write(f"done ({count})")

    def do_bootstrap(self, model_class, bootstrap_method, **bootstrap_kw):
        field_names = AuditEvent.field_names(model_class)
        if not field_names:
            raise CommandError(
                f"invalid fields ({field_names!r}) for model: {model_class}"
            )
        return bootstrap_method(
            model_class,
            field_names,
            batch_size=self.batch_size,
            **bootstrap_kw,
        )

    operations = {
        "init": init_all,
        "top-up": top_up_missing,
    }

    @contextmanager
    def bootstrap_action_log(self, *args, **kw):
        end = kw.pop("end", "\n")
        self.log_info(*args, **kw, end="")
        yield self.logfile
        self.logfile.write(end)

    def log_info(self, *args, **kw):
        self._log("INFO", *args, **kw)

    def _log(self, level_name, message, *msg_args, end="\n"):
        print(f"{level_name}: {message % msg_args}", file=self.logfile, end=end)


Command.setup_models()


class InvalidModelState(CommandError):
    pass



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('field_audit', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='auditevent',
            name='is_bootstrap',
            field=models.BooleanField(default=False),
        ),
        migrations.AddConstraint(
            model_name='auditevent',
            constraint=models.CheckConstraint(
                name='field_audit_auditevent_chk_create_or_delete_or_bootstrap',
                check=~(
                    models.Q(is_create=True, is_delete=True) | \
                    models.Q(is_create=True, is_bootstrap=True) | \
                    models.Q(is_delete=True, is_bootstrap=True)  # noqa: E502
                ),
            ),
        ),
        migrations.RemoveConstraint(
            model_name='auditevent',
            name='field_audit_auditevent_valid_create_or_delete',
        ),
    ]
